from typing import Iterable


def human_bytes(
        n: int,
        *,
        base: int = 1000,
        decimals: int = 0,
        units: Iterable[str] = ("b", "kb", "mb", "gb", "tb", "pb", "eb", "zb", "yb")
) -> str:
    """
    Convert a byte count to a human-readable string.

    Args:
        n: Byte count (e.g., from os.stat().st_size).
        base: 1000 for SI (kb, mb, ...), 1024 for IEC-like step size.
        decimals: Decimal places for non-byte units (0 -> '66kb', 1 -> '1.5gb').
        units: Unit suffixes to use. Defaults to lowercase ('kb'); swap for ('B','kB','MB',...) if preferred.

    Returns:
        A compact string like '66kb', '1mb', '1.5gb', or '999b'.
    """
    if n < 0:
        raise ValueError("Byte size cannot be negative")

    i = 0
    max_i = len(tuple(units)) - 1
    while n >= base and i < max_i:
        n /= base
        i += 1

    if i == 0 or decimals == 0:
        # Bytes or integer formatting requested
        return f"{int(n if i else n)}{tuple(units)[i]}"

    return f"{n:.{decimals}f}{tuple(units)[i]}".rstrip("0").rstrip(".")



import argparse
import logging
import platform

from pathlib import Path

from PySide6 import QtWidgets

from pixmatch.gui import MainWindow


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Process zero or more file paths."
    )
    parser.add_argument(
        "folders",
        nargs="*",
        type=Path,
        help="Folders to load into the selected file path display (to speed up testing).",
    )
    parser.add_argument('--verbose', action='store_true', help="More detailed logging")
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.verbose else logging.INFO,
        format='%(module)s::%(funcName)s::%(lineno)d %(levelname)s %(asctime)s - %(message)s',
    )

    if platform.system() == "Windows":
        # Need to tell Windows to not use the Python app icon and use the Window icon isntead...
        #    I'm not sure on the specifics but calling this method with any string seems to do the trick....
        # https://stackoverflow.com/questions/1551605
        import ctypes
        ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID('company.app.1')

    app = QtWidgets.QApplication([])
    # Basic stylesheet for subtle polish without complexity.
    app.setStyleSheet(
        """
        QToolBar { spacing: 8px; }
        QLabel#GroupTitle { padding: 4px 0; }
        QFrame#ImageTile { border: 1px solid #444; border-radius: 6px; padding: 6px; }
        """
    )
    w = MainWindow(args.folders)
    w.show()
    app.exec()



from datetime import datetime, timezone
from enum import Enum, auto
from functools import cache, lru_cache
from pathlib import Path
from typing import Dict, Iterable, List, Sequence
from zipfile import ZipFile

from PySide6 import QtCore, QtGui, QtWidgets

from pixmatch import ZipPath
from pixmatch.gui.utils import NO_MARGIN, MAX_SIZE_POLICY
from pixmatch.utils import human_bytes

ZIP_ICON_PATH = Path(__file__).resolve().parent / 'zip.png'


class SelectionState(Enum):
    """Per-thumbnail action state."""
    KEEP = auto()
    DELETE = auto()
    IGNORE = auto()


STATE_ORDER = [SelectionState.KEEP, SelectionState.DELETE, SelectionState.IGNORE]
STATE_COLORS = {
    SelectionState.KEEP: QtGui.QColor(80, 200, 120),     # green
    SelectionState.DELETE: QtGui.QColor(230, 80, 80),    # red
    SelectionState.IGNORE: QtGui.QColor(240, 190, 60),   # amber
}


# region Image view panel
def _load_pixmap(path: ZipPath, thumb_size: int) -> QtGui.QPixmap:
    """Load an image from disk and scale to a square thumbnail."""
    if path.subpath:
        with ZipFile(path.path) as zf:
            pm = QtGui.QPixmap()
            pm.loadFromData(zf.read(path.subpath))
    else:
        pm = QtGui.QPixmap(str(path.path))

    if pm.isNull():
        # Fallback: generate a checkerboard if load failed.
        pm = QtGui.QPixmap(thumb_size, thumb_size)
        pm.fill(QtGui.QColor("lightgray"))
        p = QtGui.QPainter(pm)
        p.setPen(QtCore.Qt.PenStyle.NoPen)
        c1 = QtGui.QColor(210, 210, 210)
        c2 = QtGui.QColor(180, 180, 180)
        for y in range(0, thumb_size, 16):
            for x in range(0, thumb_size, 16):
                p.setBrush(c1 if ((x // 16 + y // 16) % 2 == 0) else c2)
                p.drawRect(x, y, 16, 16)
        p.end()
    return pm.scaled(thumb_size, thumb_size,
                     QtCore.Qt.AspectRatioMode.IgnoreAspectRatio,
                     QtCore.Qt.TransformationMode.FastTransformation)


def movie_size(movie: QtGui.QMovie):
    movie.jumpToFrame(0)
    rect = QtCore.QRect()
    for i in range(movie.frameCount()):
        movie.jumpToNextFrame()
        rect |= movie.frameRect()
    width = rect.x() + rect.width()
    height = rect.y() + rect.height()

    return QtCore.QSize(width, height)


def movie_uncompressed_filesize(movie: QtGui.QMovie):
    file_size = 0
    for i in range(movie.frameCount()):
        movie.jumpToNextFrame()
        img = movie.currentImage()
        file_size += img.sizeInBytes()
    return file_size


class ImageViewPane(QtWidgets.QWidget):
    """Container with a stacked image viewer and a bottom overlay status label."""
    def __init__(self, parent=None):
        super().__init__(parent)

        # --- viewers ---
        self.current_path = None
        self._buffer = self._qbytearray = None
        self.scaled = ScaledLabel(contentsMargins=NO_MARGIN, sizePolicy=MAX_SIZE_POLICY)
        self.scaled.setMinimumSize(10, 10)

        self.raw_label = QtWidgets.QLabel()
        self.raw_label.setContentsMargins(NO_MARGIN)
        self.raw_label.setMargin(0)
        self.scroll = QtWidgets.QScrollArea()
        self.scroll.setContentsMargins(NO_MARGIN)
        self.scroll.setSizePolicy(MAX_SIZE_POLICY)
        self.scroll.setWidget(self.raw_label)

        # Only one visible at a time -> use a stack
        self.stack = QtWidgets.QStackedWidget()
        self.stack.addWidget(self.scaled)   # index 0
        self.stack.addWidget(self.scroll)   # index 1

        # --- overlay status label ---
        self.status = QtWidgets.QLabel(text="Ready", alignment=QtCore.Qt.AlignmentFlag.AlignBottom)
        self.status.setContentsMargins(NO_MARGIN)
        self.status.setObjectName("imageStatus")
        self.status.setMaximumHeight(16)
        self.status.setStyleSheet("""
            QLabel#imageStatus {
                font-weight: bold;
            }
        """)

        lay = QtWidgets.QVBoxLayout(self)
        lay.setContentsMargins(NO_MARGIN)
        lay.addWidget(self.stack)
        lay.addWidget(self.status)

    # Optional helper you can call to update the text
    def set_status(self, text: str):
        self.status.setText(text)

    def set_index(self, index: int):
        if index not in (0, 1):
            raise ValueError('Valid index must be 0 or 1 for the image pane to select!')

        self.stack.setCurrentIndex(index)

        if self.current_path:
            self.set_image(self.current_path)

        self.update()

    def clear(self):
        """Clear and reset the current object, and the two sub-objects"""
        existing_movie = self.raw_label.movie()
        if existing_movie:
            existing_movie.stop()
            existing_movie.deleteLater()
        self.raw_label.clear()

        self.scaled.clear()

    @lru_cache(maxsize=5)
    def get_movie(self, path: ZipPath):
        file_size = modified = None
        # We're setting a movie...
        if path.subpath:
            # Need to load movie from a zipfile
            with ZipFile(path.path) as zf:
                st = zf.getinfo(path.subpath)
                modified = st.date_time
                file_size = st.file_size
                self._qbytearray = QtCore.QByteArray(zf.read(path.subpath))
                self._buffer = QtCore.QBuffer(self._qbytearray)
                self._buffer.open(QtCore.QIODevice.OpenModeFlag.ReadOnly)

                movie = QtGui.QMovie()
                movie.setDevice(self._buffer)
        else:
            # Basic movie path
            movie = QtGui.QMovie(str(path.path))

        return movie, file_size, modified

    @lru_cache(maxsize=10)
    def get_pixmap(self, path: ZipPath):
        file_size = modified = None
        # We're setting an image...
        if path.subpath:
            # Need to load image from a zipfile
            with ZipFile(path.path) as zf:
                st = zf.getinfo(path.subpath)
                modified = st.date_time
                file_size = st.file_size
                pixmap = QtGui.QPixmap()
                pixmap.loadFromData(zf.read(path.subpath))
        else:
            # Basic image path
            pixmap = QtGui.QPixmap(str(path.path))

        return pixmap, file_size, modified

    def set_image(self, path: ZipPath):
        if path == self.current_path:
            return

        self.current_path = path
        file_size = modified = None
        extra = ''
        self.clear()
        if path.is_gif:
            movie, file_size, modified = self.get_movie(path)
            object_size = movie_size(movie)

            if self.stack.currentIndex() == 0:
                self.scaled.setMovie(movie)
            else:
                self.raw_label.setMovie(movie)

            uncompressed_size = movie_uncompressed_filesize(movie)

            # WEBP files which aren't animated will appear as movies with a single frame
            #   Thats boring. For the purposes of the statusbar, just treat them as images
            # TODO: Gee, that makes me wonder...
            #   could any image be treated as a movie and we could do away with this whole pixmap or movie thing?
            #   Worth investigating when I have more time...
            if movie.frameCount() > 1:
                extra = f'({human_bytes(uncompressed_size)}, {movie.frameCount()}) '
            else:
                extra = f'({human_bytes(uncompressed_size)}) '
            movie.start()
        else:
            pixmap, file_size, modified = self.get_pixmap(path)
            extra = f'({human_bytes(pixmap.toImage().sizeInBytes())}) '
            object_size = pixmap.size()

            if self.stack.currentIndex() == 0:
                self.scaled.setPixmap(pixmap)
            else:
                self.raw_label.setPixmap(pixmap)

        if self.stack.currentIndex() == 1:
            self.raw_label.resize(object_size)

        self.update()

        # region Update status text
        if not path.subpath:
            path = Path(path.path)
            st = path.stat()
            modified = datetime.fromtimestamp(st.st_mtime, tz=timezone.utc).strftime('%m/%d/%Y')
            file_size = st.st_size
        elif modified:
            modified = f"{modified[1]}/{modified[2]}/{modified[0]}"
        self.status.setText(
            f"{path.absolute()} ("
            f"{human_bytes(file_size)} {extra}"
            f"- {object_size.width()},{object_size.height()}px "
            f"- {modified}"
            f")"
        )
        # endregion


class ScaledLabel(QtWidgets.QLabel):
    """
    A version of the above ScaledLabel but for gifs/movies

    https://stackoverflow.com/questions/72188903
    https://stackoverflow.com/questions/77602181
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._movieSize = QtCore.QSize()
        self._minSize = QtCore.QSize()
        self.object_size = QtCore.QSize()
        self.orig_pixmap = self.pixmap()
        self.orig_movie = self.movie()

    def clear(self):
        super().clear()
        self.orig_pixmap = None
        if self.orig_movie:
            self.orig_movie.device().close()
            self.orig_movie.stop()
            self.orig_movie.deleteLater()
            self.orig_movie = None

    def minimumSizeHint(self):
        if self._minSize.isValid():
            return self._minSize
        return super().minimumSizeHint()

    def setPixmap(self, pixmap):  # overiding setPixmap
        if not pixmap:
            return
        self.clear()
        self.orig_pixmap = pixmap
        return super().setPixmap(self.orig_pixmap.scaled(self.frameSize(), QtCore.Qt.AspectRatioMode.KeepAspectRatio))

    def setMovie(self, movie):
        if self.movie() == movie:
            return
        if self.orig_movie and movie:
            self.clear()
        super().setMovie(movie)
        self.orig_movie = movie

        if not isinstance(movie, QtGui.QMovie) or not movie.isValid():
            self._movieSize = QtCore.QSize()
            self._minSize = QtCore.QSize()
            self.updateGeometry()
            return

        cf = movie.currentFrameNumber()
        movie.jumpToFrame(0)
        self._movieSize = movie_size(movie)
        width = self._movieSize.width()
        height = self._movieSize.height()

        minimum = min(width, height)
        maximum = max(width, height)
        ratio = maximum / minimum
        base = min(4, minimum)
        self._minSize = QtCore.QSize(base, round(base * ratio))
        if minimum == width:
            self._minSize.transpose()

        movie.jumpToFrame(cf)
        self.updateGeometry()

    def paintEvent(self, event):
        movie = self.movie()
        if not isinstance(movie, QtGui.QMovie) or not movie.isValid():
            super().paintEvent(event)
            if self.orig_pixmap:
                self.setPixmap(self.orig_pixmap)
            return

        qp = QtGui.QPainter(self)
        self.drawFrame(qp)

        cr = self.contentsRect()
        margin = self.margin()
        cr.adjust(margin, margin, -margin, -margin)

        style = self.style()
        alignment = style.visualAlignment(self.layoutDirection(), self.alignment())
        maybeSize = self._movieSize.scaled(cr.size(), QtCore.Qt.AspectRatioMode.KeepAspectRatio)

        if maybeSize != movie.scaledSize():
            movie.setScaledSize(maybeSize)
            style.drawItemPixmap(
                qp, cr, alignment,
                movie.currentPixmap().scaled(cr.size(), QtCore.Qt.AspectRatioMode.KeepAspectRatio)
            )

        else:
            style.drawItemPixmap(
                qp, cr, alignment,
                movie.currentPixmap()
            )
# endregion


# region Thumbnail tile panel
@cache
def get_overlay_icon(height, width):
    return QtGui.QPixmap(ZIP_ICON_PATH).scaled(
        int(height), int(width),
        QtCore.Qt.AspectRatioMode.KeepAspectRatio,
        QtCore.Qt.TransformationMode.FastTransformation,
    )


class ThumbnailTile(QtWidgets.QFrame):
    """
    Clickable thumbnail tile that cycles between KEEP → DELETE → IGNORE.

    Attributes:
        path: Image path (opaque identifier for the caller).
        stateChanged(path: str, state: SelectionState): Emitted on state updates.
    """
    stateChanged = QtCore.Signal(ZipPath, SelectionState)
    hovered = QtCore.Signal(ZipPath)

    def __init__(self, path: ZipPath, pixmap: QtGui.QPixmap, thumb_size: int = 32, parent=None):
        super().__init__(parent, frameShape=QtWidgets.QFrame.Shape.Box, lineWidth=2)
        self.setObjectName("ThumbnailTile")
        self.setCursor(QtCore.Qt.CursorShape.PointingHandCursor)

        self._path = path
        self._state = SelectionState.KEEP
        self._thumb_size = thumb_size

        self._image = QtWidgets.QLabel(alignment=QtCore.Qt.AlignmentFlag.AlignCenter, pixmap=pixmap)
        self._image.setFixedSize(thumb_size, thumb_size)

        lay = QtWidgets.QVBoxLayout(self)
        lay.setContentsMargins(NO_MARGIN)
        lay.setSpacing(0)
        lay.addWidget(self._image)

        if path.subpath:
            _overlay_icon = QtWidgets.QLabel(self._image)  # child of the tile so it floats over the image
            _overlay_icon.setObjectName("LockOverlay")
            _overlay_icon.setAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
            _overlay_icon.setAttribute(QtCore.Qt.WidgetAttribute.WA_TransparentForMouseEvents, True)
            _overlay_icon.setFixedSize(thumb_size, thumb_size)  # small badge; adjust later if you want
            _overlay_icon.setPixmap(get_overlay_icon(thumb_size / 1.5, thumb_size / 1.5))

        self.context_menu = QtWidgets.QMenu(self)

        act_delete = self.context_menu.addAction("Delete")
        act_ignore = self.context_menu.addAction("Ignore")
        act_ignore_group = self.context_menu.addAction("Ignore group")
        self.context_menu.addSeparator()
        act_rename = self.context_menu.addAction("Rename this file...")
        act_move = self.context_menu.addAction("Move this file")
        act_symlink = self.context_menu.addAction("Symlink this file")
        self.context_menu.addSeparator()
        act_unmark = self.context_menu.addAction("Unmark")

        # Enablement: only these three should work right now
        # If the path is from a zip (locked), disable Delete here as well.
        act_delete.setEnabled(not bool(self._path.subpath))
        act_ignore.setEnabled(True)
        act_unmark.setEnabled(True)

        # Everything else disabled for now
        act_ignore_group.setEnabled(False)
        act_rename.setEnabled(False)
        act_move.setEnabled(False)
        act_symlink.setEnabled(False)

        # Wire up state changes
        act_delete.triggered.connect(lambda _=False: setattr(self, "state", SelectionState.DELETE))
        act_ignore.triggered.connect(lambda _=False: setattr(self, "state", SelectionState.IGNORE))
        act_unmark.triggered.connect(lambda _=False: setattr(self, "state", SelectionState.KEEP))

        self._apply_state_style()

    @property
    def path(self) -> ZipPath:
        return self._path

    @property
    def state(self) -> SelectionState:
        return self._state

    @state.setter
    def state(self, state: SelectionState) -> None:
        """Set the tile selection state without cycling."""
        if self._state is state:
            return
        self._state = state
        self._apply_state_style()
        self.stateChanged.emit(self._path, self._state)

    def cycle_state(self) -> None:
        """Advance KEEP → DELETE → IGNORE → KEEP."""
        idx = STATE_ORDER.index(self._state)
        locked = bool(self._path.subpath)
        next_state = STATE_ORDER[(idx + 1) % len(STATE_ORDER)]
        if locked and next_state == SelectionState.DELETE:
            next_state = STATE_ORDER[(idx + 2) % len(STATE_ORDER)]
        self.state = next_state

    def mousePressEvent(self, e: QtGui.QMouseEvent) -> None:
        if e.button() == QtCore.Qt.MouseButton.LeftButton:
            self.cycle_state()
            e.accept()
        else:
            super().mousePressEvent(e)

    def enterEvent(self, e: QtGui.QEnterEvent) -> None:
        # fire when the cursor enters the tile
        self.hovered.emit(self._path)
        super().enterEvent(e)

    def _apply_state_style(self) -> None:
        color = STATE_COLORS[self._state]
        self.setStyleSheet(
            f"""
            QFrame#ThumbnailTile {{
                border: 2px solid rgba({color.red()}, {color.green()}, {color.blue()}, 220);
                border-radius: 6px;
                background: #202020;
            }}
            QLabel#StateBadge {{
                color: black;
                background: rgba({color.red()}, {color.green()}, {color.blue()}, 220);
                border-radius: 6px;
                font-weight: 600;
            }}
            """
        )

    def contextMenuEvent(self, event: QtGui.QContextMenuEvent) -> None:
        self.context_menu.exec(event.globalPos())
        event.accept()


class DuplicateGroupRow(QtWidgets.QWidget):
    """
    A single row of thumbnails representing one duplicate group.

    Signals:
        tileStateChanged(path: str, state: SelectionState)
    """
    tileStateChanged = QtCore.Signal(ZipPath, SelectionState)
    tileHovered = QtCore.Signal(ZipPath)

    def __init__(self, images: Sequence[ZipPath], thumb_size: int = 32, parent=None):
        super().__init__(parent)
        self._tiles: List[ThumbnailTile] = []
        self._thumb_size = thumb_size
        self.layout = QtWidgets.QHBoxLayout(self)
        self.layout.setContentsMargins(NO_MARGIN)
        self.layout.setSpacing(0)

        for path in images:
            self.add_tile(path)

        self.layout.addStretch(1)

    def tiles(self) -> Iterable[ThumbnailTile]:
        return list(self._tiles)

    def add_tile(self, path: ZipPath):
        pm = _load_pixmap(path, self._thumb_size)
        tile = ThumbnailTile(path=path, pixmap=pm, thumb_size=self._thumb_size)
        tile.stateChanged.connect(self.tileStateChanged)
        tile.hovered.connect(self.tileHovered)
        self._tiles.append(tile)
        self.layout.insertWidget(len(self._tiles) - 1, tile)


class DuplicateGroupList(QtWidgets.QWidget):
    """
    Scrollable list of duplicate groups. Each group renders as a row of thumbnails.

    Public API:
        set_groups(groups): Load groups; each group is a list of image paths.
        decisions(): Dict[path, SelectionState] for all tiles.
        set_max_rows(n): Limit how many groups to show (default 25).
        set_thumb_size(px): Set square thumbnail size (default 128).
        reset_states(): Set all tiles to KEEP.

    Notes:
        - Clicking a thumbnail cycles KEEP → DELETE → IGNORE.
        - Borders/badges are colored by state.
    """

    groupTileStateChanged = QtCore.Signal(ZipPath, SelectionState)  # path, state
    groupTileHovered = QtCore.Signal(ZipPath)

    def __init__(self, parent=None, *, max_rows: int = 25, thumb_size: int = 64, **kwargs):
        super().__init__(parent, **kwargs)
        self._max_rows = max_rows
        self._thumb_size = thumb_size

        self._scroll = QtWidgets.QScrollArea(widgetResizable=True)
        self._container = QtWidgets.QWidget()
        self._vbox = QtWidgets.QVBoxLayout(self._container)
        self._vbox.setContentsMargins(NO_MARGIN)
        self._vbox.setSpacing(0)
        _tail_spacer = QtWidgets.QSpacerItem(
            0, 0, QtWidgets.QSizePolicy.Policy.Minimum, QtWidgets.QSizePolicy.Policy.Expanding
        )
        self._vbox.addItem(_tail_spacer)
        self._scroll.setWidget(self._container)

        # Header with quick-actions.
        # self._header = QtWidgets.QHBoxLayout(contentsMargins=NO_MARGIN)
        # self._btn_keep_all = QtWidgets.QPushButton("Mark All Keep", contentsMargins=NO_MARGIN)
        # self._btn_delete_all = QtWidgets.QPushButton("Mark All Delete", contentsMargins=NO_MARGIN)
        # self._btn_ignore_all = QtWidgets.QPushButton("Mark All Ignore", contentsMargins=NO_MARGIN)
        # self._header.addWidget(self._btn_keep_all)
        # self._header.addWidget(self._btn_delete_all)
        # self._header.addWidget(self._btn_ignore_all)
        # self._header.addStretch(1)

        # self._btn_keep_all.clicked.connect(lambda: self._bulk_set(SelectionState.KEEP))
        # self._btn_delete_all.clicked.connect(lambda: self._bulk_set(SelectionState.DELETE))
        # self._btn_ignore_all.clicked.connect(lambda: self._bulk_set(SelectionState.IGNORE))

        # Main layout
        outer = QtWidgets.QVBoxLayout(self)
        outer.setContentsMargins(NO_MARGIN)
        outer.setSpacing(0)
        # self._outer.addLayout(self._header)
        outer.addWidget(self._scroll)

        self._rows: List[DuplicateGroupRow] = []

        # Status bar
        _status = QtWidgets.QHBoxLayout()
        self.first_page = QtWidgets.QPushButton("<<")
        self.page_down = QtWidgets.QPushButton("<")
        self.page_indicator = QtWidgets.QLabel(alignment=QtCore.Qt.AlignmentFlag.AlignCenter)
        self.page_up = QtWidgets.QPushButton(">")
        self.last_page = QtWidgets.QPushButton(">>")
        _status.addWidget(self.first_page)
        _status.addWidget(self.page_down)
        _status.addWidget(self.page_indicator)
        _status.addWidget(self.page_up)
        _status.addWidget(self.last_page)
        outer.addLayout(_status)

        self.update_page_indicator(1, 1)

    def set_max_rows(self, n: int) -> None:
        """Set maximum visible rows (groups)."""
        self._max_rows = max(1, int(n))

    def set_thumb_size(self, px: int) -> None:
        """Set square thumbnail size for subsequent loads."""
        self._thumb_size = max(32, int(px))

    def set_groups(self, groups: Sequence[Sequence[ZipPath]]) -> None:
        """
        Load duplicate groups.

        Args:
            groups: An iterable of groups; each group is an iterable of image file paths.
                    Only the first `max_rows` groups are shown.
        """
        self._clear_rows()
        for group in groups[: self._max_rows]:
            self.add_group(group)

    def update_page_indicator(self, current_page, total_pages):
        self.page_indicator.setText(f"Page {current_page} of {total_pages}")

    def add_group(self, group: Sequence[ZipPath]) -> None:
        if len(self._rows) == self._max_rows:
            raise ValueError("Cannot add a new group to a fully filled group list!")

        row = DuplicateGroupRow(group, thumb_size=self._thumb_size)
        row.tileStateChanged.connect(self.groupTileStateChanged)
        row.tileHovered.connect(self.groupTileHovered)
        tail_index = self._vbox.count() - 1
        self._vbox.insertWidget(tail_index, row)
        self._rows.append(row)

    def decisions(self) -> Dict[ZipPath, SelectionState]:
        """Collect {path: state} for all tiles across all rows."""
        out: Dict[ZipPath, SelectionState] = {}
        for row in self._rows:
            for tile in row.tiles():
                out[tile.path] = tile.state
        return out

    def reset_states(self) -> None:
        """Set all tiles to KEEP."""
        for row in self._rows:
            for tile in row.tiles():
                tile.state = SelectionState.KEEP

    def _clear_rows(self) -> None:
        for row in self._rows:
            row.setParent(None)
            row.deleteLater()
        self._rows.clear()
# endregion


class DirFileSystemModel(QtWidgets.QFileSystemModel):
    def hasChildren(self, /, parent: QtCore.QModelIndex | QtCore.QPersistentModelIndex = ...):
        file_info = self.fileInfo(parent)
        _dir = QtCore.QDir(file_info.absoluteFilePath())
        return bool(_dir.entryList(self.filter()))



# TODO: Validate that users don't select overlapping paths...


import logging

from importlib.metadata import version, PackageNotFoundError
from pathlib import Path

from PySide6 import QtCore, QtGui, QtWidgets

from pixmatch import ImageMatcher, ImageMatch, NewGroup, NewMatch, Finished, ZipPath
from pixmatch.gui.utils import NO_MARGIN, MAX_SIZE_POLICY
from pixmatch.gui.widgets import DuplicateGroupList, DirFileSystemModel, ImageViewPane, SelectionState
from pixmatch.utils import human_bytes

ICON_PATH = Path(__file__).resolve().parent / 'pixmatch.ico'
print(f"Icon at {ICON_PATH} : {ICON_PATH.exists()}")

logger = logging.getLogger(__name__)


def ceildiv(a, b):
    """The opposite of floordiv, //"""
    return -(a // -b)


def project_version() -> str:
    try:
        return version("pixmatch")         # e.g. "myapp"
    except PackageNotFoundError:
        return "0.0.0+unknown"


class WorkerSignals(QtCore.QObject):
    """
    Signals from a running worker thread.

    finished
        No data

    error
        tuple (exctype, value, traceback.format_exc())

    result
        object data returned from processing, anything

    progress
        float indicating % progress
    """

    new_match = QtCore.Signal(tuple)
    new_group = QtCore.Signal(object)
    finish = QtCore.Signal()


class ProcessorThread(QtCore.QRunnable):

    def __init__(self, processor, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.processor = processor
        self.signals = WorkerSignals()

        # timer lives on the GUI thread; it polls the library queue
        self._poller = QtCore.QTimer()
        self._poller.setInterval(250)
        self._poller.timeout.connect(self._drain_events)
        self._poller.start()

    def _drain_events(self):
        while not self.processor.events.empty():
            evt = self.processor.events.get_nowait()
            if isinstance(evt, NewGroup):
                self.signals.new_group.emit(evt.group)
            elif isinstance(evt, NewMatch):
                self.signals.new_match.emit((evt.group, evt.path))
            elif isinstance(evt, Finished):
                self._poller.stop()
                self.signals.finish.emit()

    def run(self):
        self.processor.run(*self.args, **self.kwargs)


class MainWindow(QtWidgets.QMainWindow):
    """
    The main application window emulating VisiPics' primary workflow.

    Key elements:
      - Toolbar with folder selection (stub), Load Test Image, Similarity slider
      - Central area that shows one group/page at a time
      - Prev/Next page navigation
    """

    def __init__(self, start_paths=None):
        super().__init__()
        self.setWindowTitle(f"PixMatch v{project_version()}")
        self.resize(1200, 800)

        # State
        self.current_page: int = 1
        self.processor = None
        self.file_states = dict()
        self.threadpool = QtCore.QThreadPool()
        self._gui_paused = False

        # UI build
        self.build_menubar()
        self.build_central()
        self.build_statusbar()
        self.build_extra()

        for start_path in start_paths or []:
            self.selected_file_path_display.addItem(str(start_path))

        if start_paths:
            self.selected_file_path_display.setCurrentRow(0)

        self.setWindowIcon(QtGui.QIcon(QtGui.QPixmap(ICON_PATH)))

    def build_extra(self):
        self.exit_warning = QtWidgets.QMessageBox(self)
        self.exit_warning.setIcon(QtWidgets.QMessageBox.Icon.Warning)
        self.exit_warning.setWindowTitle("Close?")
        self.exit_warning.setText("Are you sure you want to quit?")
        self.exit_warning.setStandardButtons(QtWidgets.QMessageBox.StandardButton.Yes | QtWidgets.QMessageBox.StandardButton.No)
        self.exit_warning.setDefaultButton(QtWidgets.QMessageBox.StandardButton.No)
        self.exit_warning.setEscapeButton(QtWidgets.QMessageBox.StandardButton.No)
        self.exit_warning.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)

    def build_menubar(self) -> None:
        """Creates the top menu bar"""
        menu = self.menuBar()

        # region File menu
        load_project = QtGui.QAction("Load Project...", self, enabled=False)
        save_project = QtGui.QAction("Save Project...", self, enabled=False)
        exit_project = QtGui.QAction("Exit", self)
        exit_project.triggered.connect(self.on_exit)

        file_menu = menu.addMenu("&File")
        file_menu.addAction(load_project)
        file_menu.addAction(save_project)
        file_menu.addSeparator()
        file_menu.addAction(exit_project)
        # endregion

        # region Edit menu
        # TODO: When adding this, make sure not to allow marking a zip file as delete
        mark_delete = QtGui.QAction("Delete", self, enabled=False)
        mark_ignore = QtGui.QAction("Ignore", self, enabled=False)
        mark_ignore_group = QtGui.QAction("Ignore Group", self, enabled=False)
        mark_ignore_folder = QtGui.QAction("Ignore Folder", self, enabled=False)
        mark_rename = QtGui.QAction("Rename this file...", self, enabled=False)
        mark_move = QtGui.QAction("Move this file...", self, enabled=False)
        mark_symlink = QtGui.QAction("Symlink this file...", self, enabled=False)
        unmark = QtGui.QAction("Un-select", self, enabled=False)

        edit_menu = menu.addMenu("&Edit")
        edit_menu.addAction(mark_delete)
        edit_menu.addAction(mark_ignore)
        edit_menu.addAction(mark_ignore_group)
        edit_menu.addAction(mark_ignore_folder)
        edit_menu.addSeparator()
        edit_menu.addAction(mark_rename)
        edit_menu.addAction(mark_move)
        edit_menu.addAction(mark_symlink)
        edit_menu.addSeparator()
        edit_menu.addAction(unmark)
        # endregion

        # region View menu
        page_next = QtGui.QAction("Next page", self)
        page_next.triggered.connect(self.on_page_up)
        page_back = QtGui.QAction("Previous page", self)
        page_back.triggered.connect(self.on_page_down)
        self.preview_resized = QtGui.QAction("Preview resized", self, checked=True, checkable=True)
        preview_full_size = QtGui.QAction("Preview full size", self, checkable=True)

        preview_options_grp = QtGui.QActionGroup(self)
        preview_options_grp.setExclusive(True)
        preview_options_grp.addAction(preview_full_size)
        preview_options_grp.addAction(self.preview_resized)

        # TODO: I'm not sure why we would want to slow display?
        #   and show differences has never worked for me
        # mark_rename = QtGui.QAction("Slow preview display", self)
        # mark_move = QtGui.QAction("Show differences", self)

        view_menu = menu.addMenu("&View")
        view_menu.addAction(page_next)
        view_menu.addAction(page_back)
        view_menu.addSeparator()
        view_menu.addAction(self.preview_resized)
        view_menu.addAction(preview_full_size)
        # endregion

        # region Tools menu
        autoselect = QtGui.QAction("Auto-select", self)
        autoselect.setEnabled(False)  # TODO:

        tool_menu = menu.addMenu("&Tools")
        tool_menu.addAction(autoselect)
        # endregion

        # region Actions menu
        run_move = QtGui.QAction("Move", self, enabled=False)
        run_delete = QtGui.QAction("Delete", self)
        run_delete.triggered.connect(self.on_delete)
        run_ignore = QtGui.QAction("Save ignored pictures", self)
        run_ignore.triggered.connect(self.on_ignore)

        actions_menu = menu.addMenu("&Actions")
        actions_menu.addAction(run_move)
        actions_menu.addAction(run_delete)
        actions_menu.addAction(run_ignore)
        # endregion

        # region Options menu
        option_hidden_folders = QtGui.QAction("Show hidden folders", self, checkable=True, enabled=False)
        option_subfolders = QtGui.QAction("Include subfolders", self, checkable=True, checked=True, enabled=False)
        option_rotations = QtGui.QAction("Scan for rotations", self, checkable=True, checked=True, enabled=False)

        # TODO: I'm not sure what these two do... I'm willing to add them if someone needs them but I don't.
        # ... = QtGui.QAction("Between folders only", self)
        # ... = QtGui.QAction("Loosen filter automatically", self)

        options_menu = menu.addMenu("&Options")
        options_menu.addAction(option_hidden_folders)
        options_menu.addAction(option_subfolders)
        options_menu.addSeparator()
        options_menu.addAction(option_rotations)
        # endregion

    def build_central(self) -> None:
        """Creates the central stacked widget area for group pages."""
        # TODO: Add other setting tabs
        style = QtWidgets.QApplication.instance().style()

        # region General controls area (top-right)
        # region Control buttons
        autoselect_btn = QtWidgets.QPushButton("Auto-select")
        autoselect_btn.setEnabled(False)  # TODO:

        tools = QtWidgets.QVBoxLayout()
        tools.addWidget(autoselect_btn)

        tool_box = QtWidgets.QGroupBox("Tools")
        tool_box.setLayout(tools)
        tool_box.setMaximumHeight(60)

        move_btn = QtWidgets.QPushButton("Move")
        move_btn.setEnabled(False)  # TODO:
        delete_btn = QtWidgets.QPushButton("Delete")
        delete_btn.pressed.connect(self.on_delete)
        save_ignored_btn = QtWidgets.QPushButton("Save ignored")
        save_ignored_btn.pressed.connect(self.on_ignore)

        actions = QtWidgets.QVBoxLayout()
        actions.addWidget(move_btn)
        actions.addWidget(delete_btn)
        actions.addWidget(save_ignored_btn)

        actions_box = QtWidgets.QGroupBox("Actions")
        actions_box.setLayout(actions)
        actions_box.setMaximumHeight(20 + 40 * 3)

        general_controls_btns = QtWidgets.QVBoxLayout()
        general_controls_btns.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        general_controls_btns.setContentsMargins(NO_MARGIN)
        general_controls_btns.addWidget(tool_box)
        general_controls_btns.addWidget(actions_box)
        # endregion

        # region Run Controls
        self.stop_btn = QtWidgets.QPushButton("\u25A0")
        self.stop_btn.setSizePolicy(MAX_SIZE_POLICY)
        self.stop_btn.setCheckable(True)
        self.stop_btn.setChecked(True)
        self.stop_btn.setStyleSheet('QPushButton {font-size: 26pt; color: maroon;}')
        self.stop_btn.clicked.connect(self.on_pause)

        self.start_btn = QtWidgets.QPushButton("\u25B6")
        self.start_btn.setSizePolicy(MAX_SIZE_POLICY)
        self.start_btn.setCheckable(True)
        self.start_btn.setStyleSheet('QPushButton {font-size: 32pt; color: green;}')
        self.start_btn.clicked.connect(self.on_start)

        self.pause_btn = QtWidgets.QPushButton()
        self.pause_btn.setSizePolicy(MAX_SIZE_POLICY)
        self.pause_btn.setCheckable(True)
        self.pause_btn.setEnabled(False)
        self.pause_btn.clicked.connect(self.on_pause)
        self.pause_btn.setIcon(style.standardIcon(
            QtWidgets.QStyle.StandardPixmap.SP_MediaPause,
        ))
        self.pause_btn.setIconSize(QtCore.QSize(28, 28))

        run_controls_options_grp = QtWidgets.QButtonGroup(self, exclusive=True)
        run_controls_options_grp.addButton(self.stop_btn)
        run_controls_options_grp.addButton(self.start_btn)
        run_controls_options_grp.addButton(self.pause_btn)

        run_controls = QtWidgets.QHBoxLayout()
        run_controls.setContentsMargins(NO_MARGIN)
        run_controls.addWidget(self.stop_btn)
        run_controls.addWidget(self.start_btn)
        run_controls.addWidget(self.pause_btn)
        # endregion

        labels = QtWidgets.QVBoxLayout()
        labels.setContentsMargins(NO_MARGIN)
        self._remaining_files_label = QtWidgets.QLabel()
        self.set_remaining_files_label(0)
        self._loaded_pictures_label = QtWidgets.QLabel()
        self.set_loaded_pictures_label(0)
        self._dup_pictures_label = QtWidgets.QLabel()
        self.set_duplicate_images_label(0)
        self._dup_groups_label = QtWidgets.QLabel()
        self.set_duplicate_groups_label(0)

        self._timer_label = QtWidgets.QLabel("00:00:00", alignment=QtCore.Qt.AlignmentFlag.AlignHCenter)
        self._elapsed_secs = 0
        self._run_timer = QtCore.QTimer(self)
        self._run_timer.setInterval(1000)  # 1s ticks
        self._run_timer.timeout.connect(self._on_run_timer_tick)

        self._label_timer = QtCore.QTimer(self)
        self._label_timer.setInterval(50)
        self._label_timer.timeout.connect(self._on_labels_tick)

        self._progress_bar = QtWidgets.QProgressBar(value=50, textVisible=False)
        labels.addWidget(self._remaining_files_label)
        labels.addWidget(self._loaded_pictures_label)
        labels.addWidget(self._dup_pictures_label)
        labels.addWidget(self._dup_groups_label)
        labels.addWidget(self._timer_label)
        labels.addWidget(self._progress_bar)
        labels.addLayout(run_controls)

        # region Settings tabs
        # region Filter tab
        slider_labels = QtWidgets.QVBoxLayout()
        slider_labels.setContentsMargins(NO_MARGIN)
        slider_labels.addWidget(QtWidgets.QLabel(text="Strict", alignment=QtCore.Qt.AlignmentFlag.AlignTop))
        slider_labels.addWidget(QtWidgets.QLabel(text="Basic", alignment=QtCore.Qt.AlignmentFlag.AlignVCenter))
        slider_labels.addWidget(QtWidgets.QLabel(text="Loose", alignment=QtCore.Qt.AlignmentFlag.AlignBottom))

        self.precision_slider = QtWidgets.QSlider(tickPosition=QtWidgets.QSlider.TickPosition.TicksLeft)
        self.precision_slider.setMaximum(10)
        self.precision_slider.setValue(5)
        self.precision_slider.sliderMoved.connect(self.on_precision_adjust)

        filter_tab_main = QtWidgets.QHBoxLayout()
        filter_tab_main.setContentsMargins(NO_MARGIN)
        filter_tab_main.addLayout(slider_labels)
        filter_tab_main.addWidget(self.precision_slider)
        filter_tab_main.addWidget(QtWidgets.QLabel(
            text="The slider determines\n"
                 "how strictly the program\n"
                 "checks for similarities\n"
                 "between the images.\n"
                 "Strict means it checks if\n"
                 "an image is the same or\n"
                 "slightly different, loose\n"
                 "allows for a greater\n"
                 "amount of differences.",
            alignment=QtCore.Qt.AlignmentFlag.AlignCenter,
        ))

        filter_tab = QtWidgets.QVBoxLayout()
        self.hash_match_chkbx = QtWidgets.QCheckBox(" Hash match")
        self.hash_match_chkbx.setEnabled(False)
        filter_tab.addWidget(self.hash_match_chkbx)
        filter_tab.addLayout(filter_tab_main)
        # endregion

        such = QtWidgets.QTabWidget()
        such.addTab(QtWidgets.QWidget(layout=filter_tab), "Filter")
        # endregion

        labels_and_such = QtWidgets.QHBoxLayout()
        labels_and_such.setContentsMargins(NO_MARGIN)
        labels_and_such.addWidget(QtWidgets.QWidget(layout=labels, maximumWidth=140))
        labels_and_such.addWidget(such)

        primary_controls = QtWidgets.QVBoxLayout()
        primary_controls.setContentsMargins(NO_MARGIN)
        primary_controls.addLayout(self.build_file_path_selection_display())
        primary_controls.addWidget(QtWidgets.QWidget(layout=labels_and_such, fixedHeight=240))

        general_controls = QtWidgets.QHBoxLayout()
        general_controls.setContentsMargins(NO_MARGIN)
        general_controls.addLayout(primary_controls)
        general_controls.addWidget(QtWidgets.QWidget(layout=general_controls_btns, maximumWidth=130))

        # region File system explorer
        file_system_model = DirFileSystemModel()
        file_system_model.setFilter(QtCore.QDir.Filter.Dirs
                                    | QtCore.QDir.Filter.Drives
                                    | QtCore.QDir.Filter.NoDotAndDotDot)
        file_system_model.setRootPath("")
        self.file_system_view = QtWidgets.QTreeView(headerHidden=True)
        self.file_system_view.setContentsMargins(NO_MARGIN)
        self.file_system_view.setModel(file_system_model)
        self.file_system_view.setRootIndex(file_system_model.index(""))
        self.file_system_view.hideColumn(1)  # Size
        self.file_system_view.hideColumn(2)  # Type
        self.file_system_view.hideColumn(3)  # Date Modified
        file_view_splitter = QtWidgets.QHBoxLayout()
        file_view_splitter.setContentsMargins(NO_MARGIN)
        file_view_splitter.addWidget(self.file_system_view)
        file_view_splitter.addWidget(QtWidgets.QWidget(layout=general_controls, maximumWidth=600))
        # endregion
        # endregion

        outer_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Horizontal, childrenCollapsible=False)
        inner_splitter = QtWidgets.QSplitter(QtCore.Qt.Orientation.Vertical, childrenCollapsible=False)

        inner_splitter.addWidget(QtWidgets.QWidget(layout=file_view_splitter))
        inner_splitter.addWidget(self.build_image_view_area())
        inner_splitter.setSizes([250, 250])

        self.duplicate_group_list = DuplicateGroupList(sizePolicy=MAX_SIZE_POLICY)
        self.duplicate_group_list.groupTileStateChanged.connect(self.on_match_state_changed)
        self.duplicate_group_list.groupTileHovered.connect(self.image_view_area.set_image)
        self.duplicate_group_list.page_down.pressed.connect(self.on_page_down)
        self.duplicate_group_list.page_up.pressed.connect(self.on_page_up)
        self.duplicate_group_list.first_page.pressed.connect(self.on_page_first)
        self.duplicate_group_list.last_page.pressed.connect(self.on_page_last)
        outer_splitter.addWidget(self.duplicate_group_list)
        outer_splitter.addWidget(inner_splitter)

        # Create a central widget and layout to hold the splitters
        central_widget = QtWidgets.QWidget()
        hbox = QtWidgets.QHBoxLayout(central_widget)
        hbox.addWidget(outer_splitter)
        self.setCentralWidget(central_widget)

    def _on_run_timer_tick(self):
        if not self.processor:
            return

        # Only count time while actively running
        if not self.processor.is_paused() and not self.processor.is_finished():
            self._elapsed_secs += 1
            h, rem = divmod(self._elapsed_secs, 3600)
            m, s = divmod(rem, 60)
            self._timer_label.setText(f"{h:02d}:{m:02d}:{s:02d}")

    def _on_labels_tick(self):
        if not self.processor:
            return

        self.update_labels()

    def on_precision_adjust(self, e):
        """The precision slider has been adjusted, so update the hash checkbox"""
        if e != 10:
            self.hash_match_chkbx.setEnabled(False)
            self.hash_match_chkbx.setChecked(False)
        else:
            self.hash_match_chkbx.setEnabled(True)

    def on_pause(self, e):
        if not e:
            return

        if not self.processor:
            return

        self.processor.pause()
        self._run_timer.stop()
        self._label_timer.stop()

    def on_start(self, e):
        if not e:
            return

        if not self.processor:
            # This is the first time running!
            self.processor = ImageMatcher(
                strength=self.precision_slider.value(),
                exact_match=self.hash_match_chkbx.isChecked(),
            )

        elif self.processor.is_paused() and not self.processor.is_finished():
            self.processor.resume()
            self.pause_btn.setEnabled(True)
            self._run_timer.start()
            self._label_timer.start()
            return

        elif self.processor.running():
            raise RuntimeError("Somehow we're trying to run when the processor appears to be running already!")

        target_paths = [
            self.selected_file_path_display.item(i).text()
            for i in range(self.selected_file_path_display.count())
        ]

        self.thread = ProcessorThread(self.processor, target_paths)
        self.thread.signals.new_group.connect(self.on_new_match_group_found)
        self.thread.signals.new_match.connect(self.on_new_match_found)
        self.thread.signals.finish.connect(self.on_finish)
        self.threadpool.start(self.thread)
        self.pause_btn.setEnabled(True)
        self.precision_slider.setEnabled(False)
        self.hash_match_chkbx.setEnabled(False)
        self._run_timer.start()
        self._label_timer.start()
        return

    def on_finish(self):
        self.stop_btn.setChecked(True)
        self.pause_btn.setEnabled(False)
        self._run_timer.stop()
        self._label_timer.stop()
        self.thread = None

    def on_delete(self, *_):
        self.process_file_states({SelectionState.DELETE})

    def on_ignore(self, *_):
        self.process_file_states({SelectionState.IGNORE})

    @property
    def last_page(self):
        return ceildiv(len(self.processor.matches), self.duplicate_group_list._max_rows)

    def on_new_match_group_found(self, match_group: ImageMatch):
        if self._gui_paused:
            return

        page_this_belongs_on, row_this_is = divmod(match_group.match_i, self.duplicate_group_list._max_rows)
        self.duplicate_group_list.update_page_indicator(self.current_page, self.last_page)

        if self.current_page == page_this_belongs_on + 1:
            self.duplicate_group_list.add_group(match_group.matches)

        self.set_duplicate_groups_label(len(self.processor.matches))

    def on_new_match_found(self, response):
        match_group: ImageMatch
        new_match: ZipPath
        match_group, new_match = response

        if self._gui_paused:
            return

        page_this_belongs_on, row_this_is = divmod(match_group.match_i, self.duplicate_group_list._max_rows)

        if self.current_page == page_this_belongs_on + 1:
            self.duplicate_group_list._rows[row_this_is].add_tile(new_match)

        self.set_duplicate_images_label(self.processor.duplicate_images)

    def set_duplicate_groups_label(self, duplicate_groups):
        self._dup_groups_label.setText(f"Duplicate groups....{duplicate_groups}")

    def set_duplicate_images_label(self, duplicate_images):
        self._dup_pictures_label.setText(f"Duplicate pictures..{duplicate_images}")

    def set_remaining_files_label(self, remaining_files):
        self._remaining_files_label.setText(f"Remaining files....{remaining_files}")

    def set_loaded_pictures_label(self, loaded_pictures):
        self._loaded_pictures_label.setText(f"Loaded pictures..{loaded_pictures}")

    def update_labels(self):
        if not self.processor:
            return

        if self.processor.found_images:
            self._progress_bar.setMaximum(self.processor.found_images)
            self._progress_bar.setValue(self.processor.processed_images)
        else:
            self._progress_bar.setMaximum(100)
            self._progress_bar.setValue(0)

        self.set_remaining_files_label(self.processor.left_to_process)
        self.set_loaded_pictures_label(self.processor.processed_images)
        self.set_duplicate_groups_label(len(self.processor.matches))
        self.set_duplicate_images_label(self.processor.duplicate_images)

    def on_exit(self, *_):
        self.close()

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        """
        Intercept window close requests (titlebar X, Cmd/Ctrl+Q, File→Exit, etc.).
        Only prompt if data is loaded.
        """
        if self.processor:
            if len(self.processor.matches) and not self.confirm_close():
                # There IS data loaded and user said NO to exiting
                event.ignore()
                return

            if self.processor.running():
                self.processor.finish()

        event.accept()

    def on_page_down(self, *_):
        if self.last_page == 1:
            # Theres only one page....
            return

        if self.current_page == 1:
            self.current_page = self.last_page
        else:
            self.current_page -= 1

        self.update_group_list()

    def on_page_up(self, *_):
        if self.last_page == 1:
            # Theres only one page....
            return

        if self.current_page == self.last_page:
            self.current_page = 1
        else:
            self.current_page += 1

        self.update_group_list()

    def on_page_first(self):
        if self.current_page != 1:
            self.current_page = 1
            self.update_group_list()

    def on_page_last(self):
        last_page = self.last_page
        if self.current_page != last_page:
            self.current_page = last_page
            self.update_group_list()

    def update_group_list(self):
        row_count = self.duplicate_group_list._max_rows
        self.duplicate_group_list.set_groups(
            [m.matches
             for m in self.processor.matches[(self.current_page - 1) * row_count:self.current_page * row_count]]
        )

        for group in self.duplicate_group_list._rows:
            for tile in group.tiles():
                set_state = self.file_states.get(tile.path)
                if set_state:
                    tile.state = set_state

        self.duplicate_group_list.update_page_indicator(self.current_page, self.last_page)

    def on_match_state_changed(self, path: ZipPath, state):
        self.file_states[path] = state

        for group in self.duplicate_group_list._rows:
            for tile in group.tiles():
                if tile.path == path:
                    tile.state = state

    # region File Path Selection display
    def build_file_path_selection_display(self):
        # TODO: I need better icons here but I can't find the "in"/"out" icons in VP execution data...

        # region Selected File Path sort controls
        file_path_up_control = QtWidgets.QPushButton("^+")
        file_path_up_control.setSizePolicy(MAX_SIZE_POLICY)
        file_path_up_control.clicked.connect(self.file_path_up_clicked)
        file_path_down_control = QtWidgets.QPushButton("V-")
        file_path_down_control.setSizePolicy(MAX_SIZE_POLICY)
        file_path_down_control.clicked.connect(self.file_path_down_clicked)

        file_path_sort_controls = QtWidgets.QVBoxLayout()
        file_path_sort_controls.setContentsMargins(NO_MARGIN)
        file_path_sort_controls.addWidget(file_path_up_control)
        file_path_sort_controls.addWidget(file_path_down_control)
        # endregion

        # region Selected File Path selection controls
        file_path_in_control = QtWidgets.QPushButton(">+")
        file_path_in_control.setSizePolicy(MAX_SIZE_POLICY)
        file_path_in_control.clicked.connect(self.file_path_in_clicked)
        file_path_out_control = QtWidgets.QPushButton("<-")
        file_path_out_control.setSizePolicy(MAX_SIZE_POLICY)
        file_path_out_control.clicked.connect(self.file_path_out_clicked)

        file_path_io_controls = QtWidgets.QVBoxLayout()
        file_path_io_controls.setContentsMargins(NO_MARGIN)
        file_path_io_controls.addWidget(file_path_in_control)
        file_path_io_controls.addWidget(file_path_out_control)
        # endregion

        self.selected_file_path_display = QtWidgets.QListWidget()

        file_path_controls = QtWidgets.QHBoxLayout()
        file_path_controls.setContentsMargins(NO_MARGIN)
        file_path_controls.addWidget(QtWidgets.QWidget(layout=file_path_io_controls, maximumWidth=50))
        file_path_controls.addWidget(self.selected_file_path_display)
        file_path_controls.addWidget(QtWidgets.QWidget(layout=file_path_sort_controls, maximumWidth=50))
        return file_path_controls

    def file_path_up_clicked(self, _):
        for selected_index in self.selected_file_path_display.selectedIndexes():
            row = selected_index.row()
            if row == 0:
                continue

            item = self.selected_file_path_display.takeItem(row)
            self.selected_file_path_display.insertItem(row - 1, item)
            self.selected_file_path_display.setCurrentIndex(self.selected_file_path_display.indexFromItem(item))

    def file_path_down_clicked(self, _):
        for selected_index in self.selected_file_path_display.selectedIndexes():
            row = selected_index.row()
            if row == self.selected_file_path_display.count():
                continue

            item = self.selected_file_path_display.takeItem(row)
            self.selected_file_path_display.insertItem(row + 1, item)
            self.selected_file_path_display.setCurrentIndex(self.selected_file_path_display.indexFromItem(item))

    def file_path_in_clicked(self, _):
        selected_indexes = self.file_system_view.selectedIndexes()
        for index in selected_indexes:
            info = self.file_system_view.model().fileInfo(index)
            target_path = info.filePath()
            self.selected_file_path_display.addItem(target_path)

            if self.processor and self.processor.running():
                self.processor.add_path(target_path)

    def file_path_out_clicked(self, _):
        for selected_index in self.selected_file_path_display.selectedIndexes():
            selected_item = self.selected_file_path_display.takeItem(selected_index.row())

            if self.processor:
                self.processor.remove_path(selected_item.text())
                self.update_group_list()
                self.update_labels()
    # endregion

    # region Image View area (bottom-right)
    def build_image_view_area(self):
        self.preview_resized.changed.connect(self.preview_resized_changed)
        self.image_view_area = ImageViewPane()
        return self.image_view_area

    def preview_resized_changed(self):
        self.image_view_area.set_index(int(not self.preview_resized.isChecked()))
    # endregion

    def build_statusbar(self) -> None:
        """Creates a simple status bar for hints and counts."""
        sb = QtWidgets.QStatusBar()
        self.setStatusBar(sb)
        self.count_label = QtWidgets.QLabel("No groups loaded")
        sb.addPermanentWidget(self.count_label)

    def process_file_states(self, states=None):
        self.image_view_area.clear()

        if not states:
            states = {SelectionState.DELETE, SelectionState.IGNORE}

        states.add(SelectionState.KEEP)

        is_paused = self.processor.is_paused()
        if not is_paused:
            self.processor.pause()

        file_size_deleted = 0
        file_count_deleted = 0
        file_count_ignored = 0
        failed_file_deletes = []

        for file, set_state in self.file_states.items():
            if set_state not in states:
                continue

            if set_state == SelectionState.DELETE:
                # TODO: This is just for testing:
                logger.info(f"Deleting {file}")
                path = Path(file.path)
                try:
                    file_size_deleted += path.stat().st_size
                    path.unlink()
                except PermissionError:
                    logger.info(f"Failed to delete {file}, it is in use!")
                    failed_file_deletes.append(file)
                else:
                    self.processor.remove(file)
                    file_count_deleted += 1
            elif set_state == SelectionState.IGNORE:
                self.processor.remove(file)
                file_count_ignored += 1
            elif set_state == SelectionState.KEEP:
                pass

        self.file_states = {
            k: v
            for k, v in self.file_states.items()
            if v not in states or k in failed_file_deletes
        }

        if self.current_page > self.last_page:
            self.current_page = self.last_page

        self.update_labels()
        self.update_group_list()
        if not is_paused:
            self.processor.resume()

        popup_text = ""

        if file_count_deleted:
            popup_text += f"Deleted {file_count_deleted} files for a savings of {human_bytes(file_size_deleted)}.\n"

        if file_count_ignored:
            popup_text += f"Removed {file_count_ignored} from matching.\n"

        if popup_text:
            dlg = QtWidgets.QMessageBox(self)
            dlg.setWindowTitle("Result")
            dlg.setText(popup_text)
            dlg.exec()

    def confirm_close(self) -> bool:
        """
        Show a confirmation dialog asking if the user really wants to close.

        Returns:
            True if the user confirmed closing; False to keep the app open.
        """
        return self.exit_warning.exec() == QtWidgets.QMessageBox.StandardButton.Yes



import hashlib
import logging
import os
import time

from collections import defaultdict
from dataclasses import dataclass, field
from multiprocessing import Pool, Manager
from pathlib import Path
from threading import Event
from typing import Union
from zipfile import ZipFile

import imagehash
import numpy as np

from PIL import Image

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ZipPath:
    path: str
    subpath: str

    @property
    def path_obj(self):
        return Path(self.path)

    @property
    def is_gif(self) -> bool:
        movie_extensions = {'.gif', '.webp'}
        return (not self.subpath and Path(self.path).suffix.lower() in movie_extensions) \
            or (self.subpath and self.subpath[-4:].lower() in movie_extensions)

    def absolute(self):
        return ZipPath(str(self.path_obj.absolute()), self.subpath)


def _is_under(folder_abs: str, target: str | Path) -> bool:
    """Return True if the ZipPath's real file (zp.path) is inside folder_abs."""
    try:
        Path(target).absolute().relative_to(Path(folder_abs).absolute())
        return True
    except ValueError:
        return False


def phash_params_for_strength(strength: int) -> tuple[int, int]:
    # TODO: This sucks.
    strength = max(0, min(10, strength))
    if strength >= 10:
        return 16, 4    # 256-bit hash, strict
    elif strength >= 8:
        return 15, 4
    elif strength >= 7:
        return 13, 4
    elif strength >= 6:
        return 11, 4
    elif strength >= 5:
        return 9, 4
    elif strength >= 4:
        return 8, 4
    elif strength >= 3:
        return 8, 3
    elif strength >= 2:
        return 7, 3
    else:
        return 6, 3


def calculate_hashes(f, is_gif=False, strength=5, exact_match=False):
    """
    Calculate hashes for a given file.

    Args:
        f (IO or str or Path): Either a file path to process, or a in-memory BytesIO object ready for reading.
        is_gif (bool): Is this gif data? Needed if passing an in-memory BytesIO object.
        strength (int): A number between 0 and 10 on the strength of the matches.
        exact_match (bool): Use exact SHA256 hahes?
            If true, strength must be 10.
            If false, perceptual hashes will be used, even with high strength.

    Returns:
        list: The found hashes.
    """
    if exact_match:
        hasher = hashlib.sha256()
        block_size = 65536
        with (open(f, "rb") if isinstance(f, (str, Path)) else f) as file:
            for block in iter(lambda: file.read(block_size), b""):
                hasher.update(block)
        return [hasher.hexdigest()]

    hash_size, highfreq_factor = phash_params_for_strength(strength)
    with (Image.open(f) as im):
        if is_gif:
            initial_hash = imagehash.phash(im, hash_size=hash_size, highfreq_factor=highfreq_factor)
            # This is going to be a bit confusing but basically, imagehash produces weird hashes for some gifs
            #   because some gifs have bad first frames consisting of nothing or only a single color...
            # To deal with that I'm looking for these bad hashes here and if its one, we advance to the next frame
            #   and use THAT for imagehash instead.
            # The ones we need to be on the lookout for are:
            #   1. The hash is all 1111...
            #   2. The hash is all 0000...
            #   3. The hash is of the form 100000...
            # TODO: This is simply not good enough. I'm still getting bad matches for gifs, tho they are extremely rare
            val = initial_hash.hash[0][0]
            while all(all(x == val for x in r) for r in initial_hash.hash) \
                    or all(all(x == np.False_ or (x_i == 0 and r_i == 0) for x_i, x in enumerate(r))
                           for r_i, r in enumerate(initial_hash.hash)):
                try:
                    im.seek(im.tell() + 1)
                except EOFError:
                    break
                else:
                    initial_hash = imagehash.phash(im, hash_size=hash_size, highfreq_factor=highfreq_factor)
                    val = initial_hash.hash[0][0]

            # For GIFs we'll look for mirrored versions but thats it
            flipped_h_image = im.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
            return [initial_hash, imagehash.phash(flipped_h_image, hash_size=hash_size, highfreq_factor=highfreq_factor)]

        flipped_h_image = im.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
        flipped_v_image = im.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
        images = (im, im.rotate(90), im.rotate(180), im.rotate(270),
                  flipped_h_image, flipped_h_image.rotate(90), flipped_h_image.rotate(180), flipped_h_image.rotate(270),
                  flipped_v_image, flipped_v_image.rotate(90), flipped_v_image.rotate(180), flipped_v_image.rotate(270))
        return [imagehash.phash(image, hash_size=hash_size, highfreq_factor=highfreq_factor) for image in images]


def _process_image(path: str | Path, strength=5, exact_match=False):
    path = Path(path)
    if path.suffix.lower() != '.zip':
        return path, calculate_hashes(path, is_gif=path.suffix.lower() in {".gif", ".webp"},
                                      strength=strength, exact_match=exact_match)

    results = dict()
    with ZipFile(path) as zf:
        for f in zf.filelist:
            with zf.open(f) as zipped_file:
                results[f.filename] = calculate_hashes(zipped_file, is_gif=f.filename[-4:].lower() in {".gif", ".webp"},
                                                       strength=strength, exact_match=exact_match)

    return path, results


@dataclass
class ImageMatch:
    match_i: int | None = field(default=None)
    matches: list[ZipPath] = field(default_factory=list)


@dataclass(frozen=True)
class NewGroup:
    group: "ImageMatch"  # forward-ref to your class


@dataclass(frozen=True)
class NewMatch:
    group: "ImageMatch"
    path: ZipPath


@dataclass(frozen=True)
class Finished:
    pass


MatcherEvent = Union[NewGroup, NewMatch, Finished]


# TODO: FINISHED signal?
class ImageMatcher:
    SUPPORTED_EXTS = {".jpg", ".jpeg", ".png", ".webp", ".tif", ".tiff", ".bmp", ".gif", ".zip"}

    def __init__(self, strength: int = 5, exact_match: bool = False, processes: int | None = None,
                 extensions: set | None = None):
        if not (0 <= strength <= 10):
            raise ValueError("Strength must be between 0 and 10!")

        self.extensions = extensions or self.SUPPORTED_EXTS

        self.strength = strength
        self.exact_match = exact_match
        self.processes = processes
        self.found_images = 0
        self.processed_images = 0
        self.duplicate_images = 0

        m = Manager()
        self.events = m.Queue()
        self._new_paths = m.Queue()
        self._removed_paths = set()
        self._processed_paths = set()
        self._hashes = defaultdict(ImageMatch)
        self._reverse_hashes = dict()

        self._not_paused = Event()
        self._not_paused.set()
        self._finished = Event()
        self._finished.set()

        self.matches = []

    def add_path(self, path: str | Path):
        path = str(Path(path).absolute())
        self._removed_paths.discard(path)
        self._new_paths.put(path)

    def remove_path(self, folder: str | Path) -> None:
        """
        Mark a folder to be skipped going forward, and remove already-indexed files
        that live under it. Pauses briefly if not already paused to keep state sane.
        """
        folder = str(Path(folder).absolute())
        paused = self.conditional_pause()
        self._removed_paths.add(folder)
        self._processed_paths.discard(folder)

        # Remove anything we've already seen under that folder
        # (iterate over a copy because remove() mutates structures)
        to_remove = [p for p in self._reverse_hashes.keys() if _is_under(folder, p.path)]
        for p in to_remove:
            self.remove(p)

        self.conditional_resume(paused)

    @property
    def left_to_process(self):
        return self.found_images - self.processed_images

    def pause(self):
        logger.debug('Performing pause')
        self._not_paused.clear()

    def conditional_pause(self):
        _conditional_pause = self.is_paused()
        if not _conditional_pause:
            logger.debug('Performing conditional pause')
            self.pause()

        return _conditional_pause

    def conditional_resume(self, was_paused):
        if not was_paused and not self.is_finished():
            logger.debug('Performing conditional resume')
            self.resume()

    def is_paused(self):
        return not self._not_paused.is_set()

    def finish(self):
        logger.debug('Performing finished')
        self._finished.set()

    def is_finished(self):
        return self._finished.is_set()

    def resume(self):
        logger.debug('Performing resume')
        self._not_paused.set()

    def running(self):
        return not self.is_paused() and (not self.is_finished() or self.left_to_process)

    def remove(self, path):
        # Pause things while we remove things...
        logger.info('Removing %s from %s', path, self.__class__.__name__)
        paused = self.conditional_pause()

        hash = self._reverse_hashes.pop(path)
        self._hashes[hash].matches.remove(path)
        if len(self._hashes[hash].matches) == 1:
            match_i = self._hashes[hash].match_i
            logger.debug('Unmatching match group %s', match_i)
            self._hashes[hash].match_i = None

            del self.matches[match_i]
            self.refresh_match_indexes(match_i)
            self.duplicate_images -= 2

        elif not self._hashes[hash].matches:
            logger.debug('Removing empty match group')
            del self._hashes[hash]

        else:
            logger.debug('Simple removal performed')
            self.duplicate_images -= 1

        self.processed_images -= 1
        self.found_images -= 1
        self.conditional_resume(paused)

    def refresh_match_indexes(self, start=0):
        for match_i, match in enumerate(self.matches[start:], start=start):
            match.match_i = match_i

    def _process_image_callback(self, result):
        self._not_paused.wait()
        if self.is_finished():
            return

        path: Path | str | ZipPath
        path, hashes = result

        if any(_is_under(d, path.path if isinstance(path, ZipPath) else path) for d in self._removed_paths):
            self.found_images -= 1
            return

        if isinstance(hashes, dict):
            for sub_path, sub_hashes in hashes.items():
                self._process_image_callback((ZipPath(str(path), sub_path), sub_hashes))
            return

        if not isinstance(path, ZipPath):
            path = ZipPath(str(path), "")

        if path in self._reverse_hashes:
            self.found_images -= 1
            return

        self.processed_images += 1
        for hash_ in hashes:
            if hash_ not in self._hashes:
                continue

            self._reverse_hashes[path] = hash_

            # This appears to be a new match!
            for match in self._hashes[hash_].matches:
                if path.absolute() == match.absolute():
                    # This appears to be a duplicate PATH...
                    logger.warning('Duplicate files entered! %s, %s', path, match)
                    return

            self._hashes[hash_].matches.append(path)
            if self._hashes[hash_].match_i is None and len(self._hashes[hash_].matches) >= 2:
                # This is a brand new match group!
                self._hashes[hash_].match_i = len(self.matches)
                self.matches.append(self._hashes[hash_])
                self.duplicate_images += 2
                self.events.put(NewGroup(self._hashes[hash_]))
                logger.debug('New match group found: %s', self._hashes[hash_].matches)
            else:
                # Just another match for an existing group...
                self.duplicate_images += 1
                self.events.put(NewMatch(self._hashes[hash_], path))
                logger.debug('New match found for group #%s: %s',
                             self._hashes[hash_].match_i,
                             self._hashes[hash_].matches)

            break
        else:
            # This is a new hash, so just add it to the hashmap and move on...
            #   Just use the initial orientation
            hash_ = hashes[0]
            self._reverse_hashes[path] = hash_
            self._hashes[hash_].matches.append(path)
            return

    def _process_image_error_callback(self, e):
        self.processed_images += 1
        print(str(e))

    def _root_stream(self):
        # Yield any paths that come up for processing, then wait until processing is finished for any new paths
        while not self._new_paths.empty() or self.left_to_process:
            if self._new_paths.empty():
                time.sleep(0.05)
                continue

            yield self._new_paths.get_nowait()

    def run(self, paths: list[str | Path]):
        # TODO: Verify none of the paths overlap
        # TODO: Verify none of the dirs have been deleted after we started

        self._not_paused.set()
        self._finished.clear()

        for path in paths:
            self.add_path(path)

        with Pool(self.processes) as tp:
            for path in self._root_stream():
                path = Path(path)
                if not path.is_dir():
                    logger.warning('A path was entered that was not a directory : %s', path)
                    continue

                path = str(path.absolute())
                if path in self._removed_paths or path in self._processed_paths:
                    continue

                for root, dirs, files in os.walk(path):
                    if self.is_finished():
                        break

                    root = Path(root)

                    if any(_is_under(d, root) for d in self._removed_paths):
                        continue

                    for f in files:
                        self._not_paused.wait()
                        if self.is_finished():
                            break

                        f = root / f

                        if f.suffix.lower() not in self.extensions:
                            continue

                        if any(_is_under(d, f) for d in self._removed_paths):
                            continue

                        # TODO: This sucks (for zips at least), but I can't iterate over the dict while its changing...
                        if ZipPath(str(f), "") in self._reverse_hashes:
                            continue

                        self.found_images += 1
                        tp.apply_async(
                            _process_image,
                            args=(f, ),
                            kwds={
                                'strength': self.strength,
                                'exact_match': self.exact_match,
                            },
                            callback=self._process_image_callback,
                            error_callback=self._process_image_error_callback,
                        )

                self._processed_paths.add(path)

            tp.close()

            if not self.is_finished():
                tp.join()

        if not self.is_finished():
            self._finished.set()
            self.events.put(Finished())



import os

from enum import Enum, auto
from typing import Dict, Iterable, List, Sequence

from PIL import Image
from PySide6 import QtCore, QtGui, QtWidgets


NO_MARGIN = QtCore.QMargins(0, 0, 0, 0)

MAX_SIZE_POLICY = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding,
                                        QtWidgets.QSizePolicy.Policy.Expanding)
